package org.vaadin.console;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.vaadin.console.Console.Command;

import com.vaadin.ui.Component;

/**
 * Java object inspector.
 *
 * With object inspector you can simply wire methods from a class as commands to
 * the Console.
 *
 * @author Sami Ekblad
 *
 */
public class ObjectInspector implements Serializable, Console.CommandProvider {
    private static final long serialVersionUID = -5001688079827306472L;
    private static final List<String> OBJECT_BLACKLIST = Arrays
            .asList(new String[] { "equals", "class", "hashCode", "notify",
                    "notifyAll", "toString", "wait" });
    private static final List<String> VAADIN_BLACKLIST = Arrays
            .asList(new String[] { "addListener", "attach", "application",
                    "changeVariables", "childRequestedRepaint",
                    "componentError", "detach", "handleError", "paint",
                    "paintContent", "removeListener", "requestRepaint",
                    "requestRepaintRequests", "style", "tag" });

    private Object theObject;
    private Map<String, Caller> commands = new HashMap<String, Caller>();
    private Set<String> ignoredCommands = new HashSet<String>();

    public ObjectInspector(Object obj) {

        // Default ignores
        ignoredCommands.addAll(OBJECT_BLACKLIST);
        if (obj instanceof Component) {
            ignoredCommands.addAll(VAADIN_BLACKLIST);
        }

        // Create bean information
        setObject(obj);

    }

    private void setObject(Object c) {
        theObject = c;
        listBeanPropertyCommands();
        listMethodCommands();
    }

    public Set<String> getAvailableCommands() {
        List<String> l = new ArrayList<String>();
        l.addAll(commands.keySet());
        Collections.sort(l);
        return new LinkedHashSet<String>(l);
    }

    public Object execute(String[] argv) throws Exception {

        if (argv == null || argv.length < 1) {
            throw new IllegalArgumentException("Missing command");
        }

        // Method name
        String method = argv[0];

        if (argv.length > 1
                && !isParamTypesOkForConsole(getCommandParams(method))) {
            throw new IllegalArgumentException("Unsupported parameter types");
        }

        Caller c = commands.get(method);
        if (c == null) {
            return null;
        }

        if (argv.length == 1 && c.isReadable()) {
            // Read
            return c.read();
        } else if (c.isWritable()) {
            // Write
            if (c.isReadable()) {
                c.write(argv);
                return c.read(); // Return the read method instead
            } else {
                return c.write(argv); // return whatever is returned
            }

        }

        return null;
    }

    private void listMethodCommands() {

        Method[] methods = theObject.getClass().getMethods();
        for (int i = 0; i < methods.length; i++) {
            final Method m = methods[i];
            if (!isBeanGetter(m) && !isBeanSetter(m)) {

                String un = m.getName();
                if (isIgnored(un)
                        || !isParamTypesOkForConsole(m.getParameterTypes())) {
                    continue;
                }

                // Make unique. Try param names first.
                if (commands.containsKey(un)) {
                    un += paramsToShortString(m.getParameterTypes());
                }

                // Fallback to appending number
                int j = 1;
                while (commands.containsKey(un)) {
                    un = m.getName() + (j++);
                }

                Caller cmd = m.getParameterTypes().length == 0 ? new Caller(m
                        .getName(), null, m.getParameterTypes()) : new Caller(
                        null, m.getName(), m.getParameterTypes());
                commands.put(un, cmd);
            }
        }
    }

    private String paramsToShortString(Class<?>[] parameterTypes) {
        String s = "";
        if (parameterTypes != null) {
            for (Class<?> c : parameterTypes) {
                s += c.getSimpleName().substring(0, 3);
            }
        }
        return s;
    }

    private boolean isBeanSetter(final Method m) {
        if (m.getName().startsWith("set")) {
            Method setMethod = null;
            try {
                String n = "g" + m.getName().substring(1);
                setMethod = m.getDeclaringClass().getMethod(n);
            } catch (SecurityException e) {
            } catch (NoSuchMethodException e) {
            }
            if (setMethod == null) {
                try {
                    String n = "is" + m.getName().substring(3);
                    setMethod = m.getDeclaringClass().getMethod(n);
                } catch (SecurityException e) {
                } catch (NoSuchMethodException e) {
                }
            }
            return setMethod != null;
        }
        return false;
    }

    private boolean isBeanGetter(Method m) {
        String n = m.getName();
        return (n.startsWith("is") && n.length() > 2)
                || (n.startsWith("get") && n.length() > 2);
    }

    /**
     * <p>
     * Perform introspection on a Java Bean class to find its properties.
     * </p>
     *
     * <p>
     * Note : This version only supports introspectable bean properties and
     * their getter and setter methods. Stand-alone <code>is</code> and
     * <code>are</code> methods are not supported.
     * </p>
     *
     * @param beanClass
     *            the Java Bean class to get properties for.
     * @return an ordered map from property names to property descriptors
     */
    static LinkedHashMap<String, PropertyDescriptor> getPropertyDescriptors(
            final Class<?> beanClass) {
        final LinkedHashMap<String, PropertyDescriptor> pdMap = new LinkedHashMap<String, PropertyDescriptor>();

        // Try to introspect, if it fails, we just have an empty Item
        try {
            final BeanInfo info = Introspector.getBeanInfo(beanClass);
            final PropertyDescriptor[] pds = info.getPropertyDescriptors();

            // Add all the bean properties as MethodProperties to this Item
            for (int i = 0; i < pds.length; i++) {
                final Method getMethod = pds[i].getReadMethod();
                if ((getMethod != null)
                        && getMethod.getDeclaringClass() != Object.class) {
                    pdMap.put(pds[i].getName(), pds[i]);
                }
            }
        } catch (final java.beans.IntrospectionException ignored) {
        }

        return pdMap;
    }

    /**
     * List commands for the theObject.
     *
     * @return
     */
    private void listBeanPropertyCommands() {

        LinkedHashMap<String, PropertyDescriptor> beanProperties = getPropertyDescriptors(theObject
                .getClass());
        // bean properties
        for (PropertyDescriptor p : beanProperties.values()) {

            // Make unique
            String name = p.getName();
            if (isIgnored(name)) {
                continue;
            }

            commands.put(name, new Caller(p.getReadMethod().getName(), p
                    .getWriteMethod() != null ? p.getWriteMethod().getName()
                    : null, p.getWriteMethod() != null ? p.getWriteMethod()
                    .getParameterTypes() : null));

            // beanCommands.put(un, p);
        }
    }

    public static boolean isParamTypesOkForConsole(Class<?>[] cls) {
        if (cls == null) {
            return true;
        }
        for (int i = 0; i < cls.length; i++) {
            if (!isParamTypeOkForConsole(cls[i])) {
                return false;
            }
        }
        return true;
    }

    public static boolean isParamTypeOkForConsole(Class<?> cls) {
        return cls.isEnum() || cls == Boolean.class || cls == boolean.class
                || cls == long.class || cls == Long.class || cls == int.class
                || cls == Integer.class || cls == byte.class
                || cls == Byte.class || cls == float.class
                || cls == Float.class || cls == double.class
                || cls == Double.class || cls == String.class;
    }

    public boolean isIgnored(String commandName) {
        return getIgnoredCommands().contains(commandName);
    }

    protected static Object[] argvToParams(String[] argv, Class<?>[] pt) {
        if (pt.length != argv.length - 1) {
            throw new IllegalArgumentException("Invalid number of parameters");
        }
        Object[] args = new Object[argv.length - 1];
        for (int j = 1; j < argv.length; j++) {
            if (pt[j - 1] == String.class) {
                args[j - 1] = argv[j];
            } else if (pt[j - 1] == Byte.class || pt[j - 1] == byte.class) {
                args[j - 1] = Byte.parseByte(argv[j]);
            } else if (pt[j - 1] == Character.class || pt[j - 1] == char.class) {
                args[j - 1] = argv[j].charAt(0);
            } else if (pt[j - 1] == Long.class || pt[j - 1] == long.class) {
                args[j - 1] = Long.parseLong(argv[j]);
            } else if (pt[j - 1] == Integer.class || pt[j - 1] == int.class) {
                args[j - 1] = Integer.parseInt(argv[j]);
            } else if (pt[j - 1] == Boolean.class || pt[j - 1] == boolean.class) {
                args[j - 1] = Boolean.parseBoolean(argv[j]);
            } else if (pt[j - 1] == Float.class || pt[j - 1] == float.class) {
                args[j - 1] = Float.parseFloat(argv[j]);
            } else if (pt[j - 1] == Double.class || pt[j - 1] == double.class) {
                args[j - 1] = Float.parseFloat(argv[j]);
            } else if (pt[j - 1].isEnum()) {
                Field e;
                try {
                    e = pt[j - 1].getDeclaredField(argv[j].toUpperCase());
                    args[j - 1] = e.get(null);
                } catch (Exception e1) {
                    throw new IllegalArgumentException("Enum not found: "
                            + argv[j].toUpperCase(), e1);
                }
            }
        }
        return args;
    }

    public Class<?>[] getCommandParams(String method) {
        if (commands.containsKey(method)) {
            Caller m = commands.get(method);
            return m.getParameterTypes() != null ? m.getParameterTypes()
                    : new Class<?>[] {};
        }
        return new Class<?>[] {};
    }

    public void setIgnoredCommands(Set<String> ignoredCommands) {
        this.ignoredCommands = ignoredCommands;
    }

    public Set<String> getIgnoredCommands() {
        return ignoredCommands;
    }

    /**
     * Method wrapper that serializes nicely.
     *
     */
    private class Caller implements Console.Command, Serializable {

        private static final long serialVersionUID = -810707579200844512L;

        private String readMethod;
        private String writeMethod;
        private Class<?>[] paramTypes;

        transient private Method rm;
        transient private Method wm;

        private Caller(String read, String write, Class<?>[] params) {
            readMethod = read;
            writeMethod = write;
            paramTypes = params;
        }

        private boolean isWritable() {
            return writeMethod != null && paramTypes != null
                    && paramTypes.length > 0;
        }

        public boolean isReadable() {
            return readMethod != null;
        }

        public Object write(Object[] params) throws SecurityException,
                NoSuchMethodException, IllegalArgumentException,
                IllegalAccessException, InvocationTargetException {
            if (wm == null) {
                wm = theObject.getClass().getMethod(writeMethod, paramTypes);
            }
            return wm.invoke(theObject, params);
        }

        public Object read() throws SecurityException, NoSuchMethodException,
                IllegalArgumentException, IllegalAccessException,
                InvocationTargetException {
            if (rm == null) {
                rm = theObject.getClass().getMethod(readMethod,
                        new Class<?>[] {});
            }
            return rm.invoke(theObject);
        }

        public Object write(String[] argv) throws SecurityException,
                NoSuchMethodException, IllegalArgumentException,
                IllegalAccessException, InvocationTargetException {
            return write(parseWriteParams(argv));
        }

        private Class<?>[] getParameterTypes() {
            return paramTypes;
        }

        private Object[] parseWriteParams(String[] argv)
                throws SecurityException, NoSuchMethodException {
            if (wm == null) {
                wm = theObject.getClass().getMethod(writeMethod, paramTypes);
            }
            return argvToParams(argv, wm.getParameterTypes());
        }

        public Object execute(Console console, String[] argv) throws Exception {
            return ObjectInspector.this.execute(argv);
        }

        public String getUsage(Console console, String[] argv) {
            return null;
        }
    }

    public String getCommandUsage(String[] argv) {
        return argv[0] + " " + paramsToString(getCommandParams(argv[0]));
    }

    private static String paramsToString(Class<?>[] paramTypes) {
        String res = "";
        if (paramTypes != null) {
            for (int i = 0; i < paramTypes.length; i++) {
                res += "<" + paramTypes[i].getSimpleName() + "> ";
            }
        }
        return res;
    }

    public Set<String> getAvailableCommands(Console console) {
        return getAvailableCommands();
    }

    public Command getCommand(Console console, String commandName) {
        return commands.get(commandName);
    }

}
